;; Generated package description from urgrep.el  -*- mode: lisp-data; no-byte-compile: t; lexical-binding:t -*-
(define-package "urgrep" "0.6.1snapshot0.20260619.3" "Universal recursive grep" '((emacs "28.1") (compat "29.1.0.1")) :commit "6db85925c7dbcc364f7264c884c13193e2ec23ec" :keywords '("grep" "search") :url "https://sr.ht/~jimporter/urgrep/")
